import graphene
from dagster import check
from dagster.core.definitions.events import AssetKey
from dagster.core.execution.backfill import BulkActionStatus
from dagster.core.host_representation import (
    InstigationSelector,
    RepositorySelector,
    ScheduleSelector,
    SensorSelector,
)
from dagster.core.scheduler.instigation import InstigatorType

from ...implementation.external import fetch_repositories, fetch_repository, fetch_workspace
from ...implementation.fetch_assets import get_asset, get_asset_node, get_asset_nodes, get_assets
from ...implementation.fetch_backfills import get_backfill, get_backfills
from ...implementation.fetch_jobs import get_job_state_or_error, get_unloadable_job_states_or_error
from ...implementation.fetch_partition_sets import get_partition_set, get_partition_sets_or_error
from ...implementation.fetch_pipelines import (
    get_pipeline_or_error,
    get_pipeline_snapshot_or_error_from_pipeline_selector,
    get_pipeline_snapshot_or_error_from_snapshot_id,
)
from ...implementation.fetch_runs import (
    get_execution_plan,
    get_run_by_id,
    get_run_group,
    get_run_groups,
    get_run_tags,
    validate_pipeline_config,
)
from ...implementation.fetch_schedules import (
    get_schedule_or_error,
    get_scheduler_or_error,
    get_schedules_or_error,
)
from ...implementation.fetch_sensors import get_sensor_or_error, get_sensors_or_error
from ...implementation.fetch_solids import get_graph_or_error
from ...implementation.run_config_schema import resolve_run_config_schema_or_error
from ...implementation.utils import graph_selector_from_graphql, pipeline_selector_from_graphql
from ..asset_graph import GrapheneAssetNode, GrapheneAssetNodeOrError
from ..backfill import (
    GrapheneBulkActionStatus,
    GraphenePartitionBackfillOrError,
    GraphenePartitionBackfillsOrError,
)
from ..external import (
    GrapheneRepositoriesOrError,
    GrapheneRepositoryOrError,
    GrapheneWorkspaceOrError,
)
from ..inputs import (
    GrapheneAssetKeyInput,
    GrapheneGraphSelector,
    GrapheneInstigationSelector,
    GraphenePipelineSelector,
    GrapheneRepositorySelector,
    GrapheneRunsFilter,
    GrapheneScheduleSelector,
    GrapheneSensorSelector,
)
from ..instance import GrapheneInstance
from ..instigation import (
    GrapheneInstigationStateOrError,
    GrapheneInstigationStatesOrError,
    GrapheneInstigationType,
)
from ..partition_sets import GraphenePartitionSetOrError, GraphenePartitionSetsOrError
from ..permissions import GraphenePermission
from ..pipelines.config_result import GraphenePipelineConfigValidationResult
from ..pipelines.pipeline import GrapheneRunOrError
from ..pipelines.snapshot import GraphenePipelineSnapshotOrError
from ..run_config import GrapheneRunConfigSchemaOrError
from ..runs import (
    GrapheneRunConfigData,
    GrapheneRunGroupOrError,
    GrapheneRunGroupsOrError,
    GrapheneRuns,
    GrapheneRunsOrError,
    parse_run_config_input,
)
from ..schedules import GrapheneScheduleOrError, GrapheneSchedulerOrError, GrapheneSchedulesOrError
from ..sensors import GrapheneSensorOrError, GrapheneSensorsOrError
from ..tags import GraphenePipelineTagAndValues
from ..util import non_null_list
from .assets import GrapheneAssetOrError, GrapheneAssetsOrError
from .execution_plan import GrapheneExecutionPlanOrError
from .pipeline import GrapheneGraphOrError, GraphenePipelineOrError


class GrapheneDagitQuery(graphene.ObjectType):
    class Meta:
        name = "DagitQuery"

    version = graphene.NonNull(graphene.String)

    repositoriesOrError = graphene.NonNull(GrapheneRepositoriesOrError)
    repositoryOrError = graphene.Field(
        graphene.NonNull(GrapheneRepositoryOrError),
        repositorySelector=graphene.NonNull(GrapheneRepositorySelector),
    )

    workspaceOrError = graphene.NonNull(GrapheneWorkspaceOrError)

    pipelineOrError = graphene.Field(
        graphene.NonNull(GraphenePipelineOrError), params=graphene.NonNull(GraphenePipelineSelector)
    )

    pipelineSnapshotOrError = graphene.Field(
        graphene.NonNull(GraphenePipelineSnapshotOrError),
        snapshotId=graphene.String(),
        activePipelineSelector=graphene.Argument(GraphenePipelineSelector),
    )

    graphOrError = graphene.Field(
        graphene.NonNull(GrapheneGraphOrError),
        selector=graphene.Argument(GrapheneGraphSelector),
    )

    scheduler = graphene.Field(graphene.NonNull(GrapheneSchedulerOrError))

    scheduleOrError = graphene.Field(
        graphene.NonNull(GrapheneScheduleOrError),
        schedule_selector=graphene.NonNull(GrapheneScheduleSelector),
    )

    schedulesOrError = graphene.Field(
        graphene.NonNull(GrapheneSchedulesOrError),
        repositorySelector=graphene.NonNull(GrapheneRepositorySelector),
    )

    sensorOrError = graphene.Field(
        graphene.NonNull(GrapheneSensorOrError),
        sensorSelector=graphene.NonNull(GrapheneSensorSelector),
    )
    sensorsOrError = graphene.Field(
        graphene.NonNull(GrapheneSensorsOrError),
        repositorySelector=graphene.NonNull(GrapheneRepositorySelector),
    )

    instigationStateOrError = graphene.Field(
        graphene.NonNull(GrapheneInstigationStateOrError),
        instigationSelector=graphene.NonNull(GrapheneInstigationSelector),
    )

    unloadableInstigationStatesOrError = graphene.Field(
        graphene.NonNull(GrapheneInstigationStatesOrError),
        instigationType=graphene.Argument(GrapheneInstigationType),
    )

    partitionSetsOrError = graphene.Field(
        graphene.NonNull(GraphenePartitionSetsOrError),
        repositorySelector=graphene.NonNull(GrapheneRepositorySelector),
        pipelineName=graphene.NonNull(graphene.String),
    )
    partitionSetOrError = graphene.Field(
        graphene.NonNull(GraphenePartitionSetOrError),
        repositorySelector=graphene.NonNull(GrapheneRepositorySelector),
        partitionSetName=graphene.String(),
    )

    pipelineRunsOrError = graphene.Field(
        graphene.NonNull(GrapheneRunsOrError),
        filter=graphene.Argument(GrapheneRunsFilter),
        cursor=graphene.String(),
        limit=graphene.Int(),
    )
    pipelineRunOrError = graphene.Field(
        graphene.NonNull(GrapheneRunOrError), runId=graphene.NonNull(graphene.ID)
    )
    runsOrError = graphene.Field(
        graphene.NonNull(GrapheneRunsOrError),
        filter=graphene.Argument(GrapheneRunsFilter),
        cursor=graphene.String(),
        limit=graphene.Int(),
    )
    runOrError = graphene.Field(
        graphene.NonNull(GrapheneRunOrError), runId=graphene.NonNull(graphene.ID)
    )
    pipelineRunTags = non_null_list(GraphenePipelineTagAndValues)

    runGroupOrError = graphene.Field(
        graphene.NonNull(GrapheneRunGroupOrError), runId=graphene.NonNull(graphene.ID)
    )

    runGroupsOrError = graphene.Field(
        graphene.NonNull(GrapheneRunGroupsOrError),
        filter=graphene.Argument(GrapheneRunsFilter),
        cursor=graphene.String(),
        limit=graphene.Int(),
    )

    isPipelineConfigValid = graphene.Field(
        graphene.NonNull(GraphenePipelineConfigValidationResult),
        args={
            "pipeline": graphene.Argument(graphene.NonNull(GraphenePipelineSelector)),
            "runConfigData": graphene.Argument(GrapheneRunConfigData),
            "mode": graphene.Argument(graphene.NonNull(graphene.String)),
        },
    )

    executionPlanOrError = graphene.Field(
        graphene.NonNull(GrapheneExecutionPlanOrError),
        args={
            "pipeline": graphene.Argument(graphene.NonNull(GraphenePipelineSelector)),
            "runConfigData": graphene.Argument(GrapheneRunConfigData),
            "mode": graphene.Argument(graphene.NonNull(graphene.String)),
        },
    )

    runConfigSchemaOrError = graphene.Field(
        graphene.NonNull(GrapheneRunConfigSchemaOrError),
        args={
            "selector": graphene.Argument(graphene.NonNull(GraphenePipelineSelector)),
            "mode": graphene.Argument(graphene.String),
        },
        description="""Fetch an environment schema given an execution selection and a mode.
        See the descripton on RunConfigSchema for more information.""",
    )

    instance = graphene.NonNull(GrapheneInstance)

    assetsOrError = graphene.Field(
        graphene.NonNull(GrapheneAssetsOrError),
        prefix=graphene.List(graphene.NonNull(graphene.String)),
        cursor=graphene.String(),
        limit=graphene.Int(),
    )

    assetOrError = graphene.Field(
        graphene.NonNull(GrapheneAssetOrError),
        assetKey=graphene.Argument(graphene.NonNull(GrapheneAssetKeyInput)),
    )

    assetNodes = non_null_list(GrapheneAssetNode)

    assetNodeOrError = graphene.Field(
        graphene.NonNull(GrapheneAssetNodeOrError),
        assetKey=graphene.Argument(graphene.NonNull(GrapheneAssetKeyInput)),
    )

    partitionBackfillOrError = graphene.Field(
        graphene.NonNull(GraphenePartitionBackfillOrError),
        backfillId=graphene.Argument(graphene.NonNull(graphene.String)),
    )

    partitionBackfillsOrError = graphene.Field(
        graphene.NonNull(GraphenePartitionBackfillsOrError),
        status=graphene.Argument(GrapheneBulkActionStatus),
        cursor=graphene.String(),
        limit=graphene.Int(),
    )

    permissions = graphene.Field(non_null_list(GraphenePermission))

    def resolve_repositoriesOrError(self, graphene_info):
        return fetch_repositories(graphene_info)

    def resolve_repositoryOrError(self, graphene_info, **kwargs):
        return fetch_repository(
            graphene_info,
            RepositorySelector.from_graphql_input(kwargs.get("repositorySelector")),
        )

    def resolve_workspaceOrError(self, graphene_info):
        return fetch_workspace(graphene_info.context)

    def resolve_pipelineSnapshotOrError(self, graphene_info, **kwargs):
        snapshot_id_arg = kwargs.get("snapshotId")
        pipeline_selector_arg = kwargs.get("activePipelineSelector")
        check.invariant(
            not (snapshot_id_arg and pipeline_selector_arg),
            "Must only pass one of snapshotId or activePipelineSelector",
        )
        check.invariant(
            snapshot_id_arg or pipeline_selector_arg,
            "Must set one of snapshotId or activePipelineSelector",
        )

        if pipeline_selector_arg:
            pipeline_selector = pipeline_selector_from_graphql(kwargs["activePipelineSelector"])
            return get_pipeline_snapshot_or_error_from_pipeline_selector(
                graphene_info, pipeline_selector
            )
        else:
            return get_pipeline_snapshot_or_error_from_snapshot_id(graphene_info, snapshot_id_arg)

    def resolve_graphOrError(self, graphene_info, **kwargs):
        graph_selector = graph_selector_from_graphql(kwargs["selector"])
        return get_graph_or_error(graphene_info, graph_selector)

    def resolve_version(self, graphene_info):
        return graphene_info.context.version

    def resolve_scheduler(self, graphene_info):
        return get_scheduler_or_error(graphene_info)

    def resolve_scheduleOrError(self, graphene_info, schedule_selector):
        return get_schedule_or_error(
            graphene_info, ScheduleSelector.from_graphql_input(schedule_selector)
        )

    def resolve_schedulesOrError(self, graphene_info, **kwargs):
        return get_schedules_or_error(
            graphene_info,
            RepositorySelector.from_graphql_input(kwargs.get("repositorySelector")),
        )

    def resolve_sensorOrError(self, graphene_info, sensorSelector):
        return get_sensor_or_error(graphene_info, SensorSelector.from_graphql_input(sensorSelector))

    def resolve_sensorsOrError(self, graphene_info, **kwargs):
        return get_sensors_or_error(
            graphene_info,
            RepositorySelector.from_graphql_input(kwargs.get("repositorySelector")),
        )

    def resolve_instigationStateOrError(self, graphene_info, instigationSelector):
        return get_job_state_or_error(
            graphene_info, InstigationSelector.from_graphql_input(instigationSelector)
        )

    def resolve_unloadableInstigationStatesOrError(self, graphene_info, **kwargs):
        job_type = (
            InstigatorType(kwargs["instigationType"]) if "instigationType" in kwargs else None
        )
        return get_unloadable_job_states_or_error(graphene_info, job_type)

    def resolve_pipelineOrError(self, graphene_info, **kwargs):
        return get_pipeline_or_error(
            graphene_info,
            pipeline_selector_from_graphql(kwargs["params"]),
        )

    def resolve_pipelineRunsOrError(self, _graphene_info, **kwargs):
        filters = kwargs.get("filter")
        if filters is not None:
            filters = filters.to_selector()

        return GrapheneRuns(
            filters=filters,
            cursor=kwargs.get("cursor"),
            limit=kwargs.get("limit"),
        )

    def resolve_pipelineRunOrError(self, graphene_info, runId):
        return get_run_by_id(graphene_info, runId)

    def resolve_runsOrError(self, _graphene_info, **kwargs):
        filters = kwargs.get("filter")
        if filters is not None:
            filters = filters.to_selector()

        return GrapheneRuns(
            filters=filters,
            cursor=kwargs.get("cursor"),
            limit=kwargs.get("limit"),
        )

    def resolve_runOrError(self, graphene_info, runId):
        return get_run_by_id(graphene_info, runId)

    def resolve_runGroupsOrError(self, graphene_info, **kwargs):
        filters = kwargs.get("filter")
        if filters is not None:
            filters = filters.to_selector()

        return GrapheneRunGroupsOrError(
            results=get_run_groups(
                graphene_info, filters, kwargs.get("cursor"), kwargs.get("limit")
            )
        )

    def resolve_partitionSetsOrError(self, graphene_info, **kwargs):
        return get_partition_sets_or_error(
            graphene_info,
            RepositorySelector.from_graphql_input(kwargs.get("repositorySelector")),
            kwargs.get("pipelineName"),
        )

    def resolve_partitionSetOrError(self, graphene_info, **kwargs):
        return get_partition_set(
            graphene_info,
            RepositorySelector.from_graphql_input(kwargs.get("repositorySelector")),
            kwargs.get("partitionSetName"),
        )

    def resolve_pipelineRunTags(self, graphene_info):
        return get_run_tags(graphene_info)

    def resolve_runGroupOrError(self, graphene_info, runId):
        return get_run_group(graphene_info, runId)

    def resolve_isPipelineConfigValid(self, graphene_info, pipeline, **kwargs):
        return validate_pipeline_config(
            graphene_info,
            pipeline_selector_from_graphql(pipeline),
            parse_run_config_input(kwargs.get("runConfigData", {})),
            kwargs.get("mode"),
        )

    def resolve_executionPlanOrError(self, graphene_info, pipeline, **kwargs):
        return get_execution_plan(
            graphene_info,
            pipeline_selector_from_graphql(pipeline),
            parse_run_config_input(kwargs.get("runConfigData", {})),
            kwargs.get("mode"),
        )

    def resolve_runConfigSchemaOrError(self, graphene_info, **kwargs):
        return resolve_run_config_schema_or_error(
            graphene_info,
            pipeline_selector_from_graphql(kwargs["selector"]),
            kwargs.get("mode"),
        )

    def resolve_instance(self, graphene_info):
        return GrapheneInstance(graphene_info.context.instance)

    def resolve_assetNodes(self, graphene_info):
        return get_asset_nodes(graphene_info)

    def resolve_assetNodeOrError(self, graphene_info, **kwargs):
        return get_asset_node(graphene_info, AssetKey.from_graphql_input(kwargs["assetKey"]))

    def resolve_assetsOrError(self, graphene_info, **kwargs):
        return get_assets(
            graphene_info,
            prefix=kwargs.get("prefix"),
            cursor=kwargs.get("cursor"),
            limit=kwargs.get("limit"),
        )

    def resolve_assetOrError(self, graphene_info, **kwargs):
        return get_asset(graphene_info, AssetKey.from_graphql_input(kwargs["assetKey"]))

    def resolve_partitionBackfillOrError(self, graphene_info, backfillId):
        return get_backfill(graphene_info, backfillId)

    def resolve_partitionBackfillsOrError(self, graphene_info, **kwargs):
        status = kwargs.get("status")
        return get_backfills(
            graphene_info,
            status=BulkActionStatus.from_graphql_input(status) if status else None,
            cursor=kwargs.get("cursor"),
            limit=kwargs.get("limit"),
        )

    def resolve_permissions(self, graphene_info):
        permissions = graphene_info.context.permissions
        return [GraphenePermission(permission, value) for permission, value in permissions.items()]



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot

snapshots = Snapshot()

snapshots['test_mode_fetch_loggers 1'] = {
    'pipelineOrError': {
        '__typename': 'Pipeline',
        'modes': [
            {
                'loggers': [
                    {
                        'configField': {
                            'configType': {
                                'fields': [
                                    {
                                        'configType': {
                                            'key': 'String'
                                        },
                                        'name': 'log_level'
                                    },
                                    {
                                        'configType': {
                                            'key': 'String'
                                        },
                                        'name': 'prefix'
                                    }
                                ],
                                'key': 'Shape.37a85e53844c27abff7868c5c2e6a32d7fd1c308'
                            }
                        },
                        'description': None,
                        'name': 'bar'
                    }
                ],
                'name': 'bar_mode'
            },
            {
                'loggers': [
                    {
                        'configField': {
                            'configType': {
                                'key': 'String'
                            }
                        },
                        'description': None,
                        'name': 'foo'
                    }
                ],
                'name': 'foo_mode'
            },
            {
                'loggers': [
                    {
                        'configField': {
                            'configType': {
                                'fields': [
                                    {
                                        'configType': {
                                            'key': 'String'
                                        },
                                        'name': 'log_level'
                                    },
                                    {
                                        'configType': {
                                            'key': 'String'
                                        },
                                        'name': 'prefix'
                                    }
                                ],
                                'key': 'Shape.37a85e53844c27abff7868c5c2e6a32d7fd1c308'
                            }
                        },
                        'description': None,
                        'name': 'bar'
                    },
                    {
                        'configField': {
                            'configType': {
                                'key': 'String'
                            }
                        },
                        'description': None,
                        'name': 'foo'
                    }
                ],
                'name': 'foobar_mode'
            }
        ]
    }
}



from dagster import PipelineDefinition, PipelineRunStatus, check
from dagster.config.validate import validate_config
from dagster.core.definitions import create_run_config_schema
from dagster.core.errors import DagsterRunNotFoundError
from dagster.core.execution.stats import StepEventStatus
from dagster.core.host_representation import PipelineSelector
from dagster.core.storage.pipeline_run import PipelineRunsFilter
from dagster.core.storage.tags import TagType, get_tag_type
from graphql.execution.base import ResolveInfo

from .external import ensure_valid_config, get_external_pipeline_or_raise
from .utils import UserFacingGraphQLError, capture_error


def is_config_valid(pipeline_def, run_config, mode):
    check.str_param(mode, "mode")
    check.inst_param(pipeline_def, "pipeline_def", PipelineDefinition)

    run_config_schema = create_run_config_schema(pipeline_def, mode)
    validated_config = validate_config(run_config_schema.config_type, run_config)
    return validated_config.success


def get_validated_config(pipeline_def, run_config, mode):
    from ..schema.pipelines.config import GrapheneRunConfigValidationInvalid

    check.str_param(mode, "mode")
    check.inst_param(pipeline_def, "pipeline_def", PipelineDefinition)

    run_config_schema = create_run_config_schema(pipeline_def, mode)

    validated_config = validate_config(run_config_schema.config_type, run_config)

    if not validated_config.success:
        raise UserFacingGraphQLError(
            GrapheneRunConfigValidationInvalid.for_validation_errors(
                pipeline_def.get_external_pipeline(), validated_config.errors
            )
        )

    return validated_config


def get_run_by_id(graphene_info, run_id):
    from ..schema.errors import GrapheneRunNotFoundError
    from ..schema.pipelines.pipeline import GrapheneRun

    instance = graphene_info.context.instance
    run = instance.get_run_by_id(run_id)
    if not run:
        return GrapheneRunNotFoundError(run_id)
    else:
        return GrapheneRun(run)


def get_run_tags(graphene_info):
    from ..schema.tags import GraphenePipelineTagAndValues

    instance = graphene_info.context.instance
    return [
        GraphenePipelineTagAndValues(key=key, values=values)
        for key, values in instance.get_run_tags()
        if get_tag_type(key) != TagType.HIDDEN
    ]


@capture_error
def get_run_group(graphene_info, run_id):
    from ..schema.errors import GrapheneRunGroupNotFoundError
    from ..schema.pipelines.pipeline import GrapheneRun
    from ..schema.runs import GrapheneRunGroup

    instance = graphene_info.context.instance
    try:
        result = instance.get_run_group(run_id)
    except DagsterRunNotFoundError:
        return GrapheneRunGroupNotFoundError(run_id)
    root_run_id, run_group = result
    return GrapheneRunGroup(root_run_id=root_run_id, runs=[GrapheneRun(run) for run in run_group])


def get_runs(graphene_info, filters, cursor=None, limit=None):
    from ..schema.pipelines.pipeline import GrapheneRun

    check.opt_inst_param(filters, "filters", PipelineRunsFilter)
    check.opt_str_param(cursor, "cursor")
    check.opt_int_param(limit, "limit")

    instance = graphene_info.context.instance
    runs = []

    if filters and filters.run_ids and len(filters.run_ids) == 1:
        run = instance.get_run_by_id(filters.run_ids[0])
        if run:
            runs = [run]
    elif filters and (
        filters.pipeline_name
        or filters.tags
        or filters.statuses
        or filters.snapshot_id
        or filters.run_ids
    ):
        runs = instance.get_runs(filters, cursor, limit)
    else:
        runs = instance.get_runs(cursor=cursor, limit=limit)

    return [GrapheneRun(run) for run in runs]


def get_in_progress_runs_for_job(graphene_info, job_name):
    instance = graphene_info.context.instance

    in_progress_runs_filter = PipelineRunsFilter(
        pipeline_name=job_name,
        statuses=[
            PipelineRunStatus.STARTING,
            PipelineRunStatus.MANAGED,
            PipelineRunStatus.NOT_STARTED,
            PipelineRunStatus.QUEUED,
            PipelineRunStatus.STARTED,
            PipelineRunStatus.CANCELING,
        ],
    )

    return instance.get_runs(in_progress_runs_filter)


def get_in_progress_runs_by_step(graphene_info, in_progress_runs, step_keys):
    from ..schema.pipelines.pipeline import GrapheneInProgressRunsByStep, GrapheneRun

    in_progress_runs_by_step = {}
    unstarted_runs_by_step = {}

    for run in in_progress_runs:
        step_stats = graphene_info.context.instance.get_run_step_stats(run.run_id, step_keys)
        for step_stat in step_stats:
            if step_stat.status == StepEventStatus.IN_PROGRESS:
                if step_stat.step_key not in in_progress_runs_by_step:
                    in_progress_runs_by_step[step_stat.step_key] = []
                in_progress_runs_by_step[step_stat.step_key].append(GrapheneRun(run))

        asset_names = graphene_info.context.instance.get_execution_plan_snapshot(
            run.execution_plan_snapshot_id
        ).step_keys_to_execute

        for step_key in asset_names:
            # step_stats only contains stats for steps that are in progress or complete
            is_unstarted = (
                len([step_stat for step_stat in step_stats if step_stat.step_key == step_key]) == 0
            )
            if is_unstarted:
                if step_key not in unstarted_runs_by_step:
                    unstarted_runs_by_step[step_key] = []
                unstarted_runs_by_step[step_key].append(GrapheneRun(run))

    step_runs = []
    for key in in_progress_runs_by_step.keys() | unstarted_runs_by_step.keys():
        step_runs.append(
            GrapheneInProgressRunsByStep(
                key,
                unstarted_runs_by_step.get(key, []),
                in_progress_runs_by_step.get(key, []),
            )
        )

    return step_runs


def get_runs_count(graphene_info, filters):
    return graphene_info.context.instance.get_runs_count(filters)


def get_run_groups(graphene_info, filters=None, cursor=None, limit=None):
    from ..schema.pipelines.pipeline import GrapheneRun
    from ..schema.runs import GrapheneRunGroup

    check.opt_inst_param(filters, "filters", PipelineRunsFilter)
    check.opt_str_param(cursor, "cursor")
    check.opt_int_param(limit, "limit")

    instance = graphene_info.context.instance
    run_groups = instance.get_run_groups(filters=filters, cursor=cursor, limit=limit)

    for root_run_id in run_groups:
        run_groups[root_run_id]["runs"] = [
            GrapheneRun(run) for run in run_groups[root_run_id]["runs"]
        ]

    return [
        GrapheneRunGroup(root_run_id=root_run_id, runs=run_group["runs"])
        for root_run_id, run_group in run_groups.items()
    ]


@capture_error
def validate_pipeline_config(graphene_info, selector, run_config, mode):
    from ..schema.pipelines.config import GraphenePipelineConfigValidationValid

    check.inst_param(graphene_info, "graphene_info", ResolveInfo)
    check.inst_param(selector, "selector", PipelineSelector)
    check.opt_str_param(mode, "mode")

    external_pipeline = get_external_pipeline_or_raise(graphene_info, selector)
    ensure_valid_config(external_pipeline, mode, run_config)
    return GraphenePipelineConfigValidationValid(pipeline_name=external_pipeline.name)


@capture_error
def get_execution_plan(graphene_info, selector, run_config, mode):
    from ..schema.execution import GrapheneExecutionPlan

    check.inst_param(graphene_info, "graphene_info", ResolveInfo)
    check.inst_param(selector, "selector", PipelineSelector)
    check.opt_str_param(mode, "mode")

    external_pipeline = get_external_pipeline_or_raise(graphene_info, selector)
    ensure_valid_config(external_pipeline, mode, run_config)
    return GrapheneExecutionPlan(
        graphene_info.context.get_external_execution_plan(
            external_pipeline=external_pipeline,
            mode=mode,
            run_config=run_config,
            step_keys_to_execute=None,
            known_state=None,
        )
    )


@capture_error
def get_stats(graphene_info, run_id):
    from ..schema.pipelines.pipeline_run_stats import GrapheneRunStatsSnapshot

    stats = graphene_info.context.instance.get_run_stats(run_id)
    stats.id = "stats-{run_id}"
    return GrapheneRunStatsSnapshot(stats)


def get_step_stats(graphene_info, run_id, step_keys=None):
    from ..schema.logs.events import GrapheneRunStepStats

    step_stats = graphene_info.context.instance.get_run_step_stats(run_id, step_keys)
    return [GrapheneRunStepStats(stats) for stats in step_stats]



import os
from enum import Enum
from threading import Event, Thread
from time import sleep

import gevent
from dagster import check


class State(Enum):
    # initialized
    NULL = 0
    # streaming the existing event log entries
    LOADING = 1
    # watching for new writes
    WATCHING = 2


def get_chunk_size() -> int:
    return int(os.getenv("DAGIT_EVENT_LOAD_CHUNK_SIZE", "10000"))


class PipelineRunObservableSubscribe:
    def __init__(self, instance, run_id, after_cursor=None):
        self.instance = instance
        self.run_id = run_id
        self.observer = None
        self.state = State.NULL
        self.stopping = None
        self.stopped = None
        self.after_cursor = after_cursor if after_cursor is not None else -1

    def __call__(self, observer):
        self.observer = observer
        check.invariant(self.state is State.NULL, f"unexpected state {self.state}")
        chunk_size = get_chunk_size()
        events = self.instance.logs_after(self.run_id, self.after_cursor, limit=chunk_size)
        done_loading = len(events) < chunk_size

        if events:
            self.observer.on_next((events, not done_loading))
            self.after_cursor = len(events) + int(self.after_cursor)

        if done_loading:
            self.watch_events()
        else:
            self.load_events()

        return self

    def load_events(self):
        self.state = State.LOADING

        # support for gevent based dagit server
        if gevent.get_hub().gr_frame:
            self.stopping = gevent.event.Event()
            self.stopped = gevent.event.Event()
            load_thread = gevent.Greenlet(self.background_event_loading, gevent.sleep)
        else:
            self.stopping = Event()
            self.stopped = Event()
            load_thread = Thread(
                target=self.background_event_loading,
                args=(sleep,),
                name=f"load-events-{self.run_id}",
            )

        load_thread.start()

    def watch_events(self):
        self.state = State.WATCHING
        self.instance.watch_event_logs(self.run_id, self.after_cursor, self.handle_new_event)

    def background_event_loading(self, sleep_fn):
        chunk_size = get_chunk_size()

        while not self.stopping.is_set():
            events = self.instance.logs_after(self.run_id, self.after_cursor, limit=chunk_size)
            if self.observer is None:
                break

            done_loading = len(events) < chunk_size

            self.observer.on_next((events, not done_loading))
            self.after_cursor = len(events) + int(self.after_cursor)

            if done_loading:
                break

            sleep_fn(0)

        if not self.stopping.is_set():
            self.watch_events()

        self.stopped.set()
        return

    def dispose(self):
        # called when the connection gets closed, allowing the observer to get GC'ed
        if self.observer and callable(getattr(self.observer, "dispose", None)):
            self.observer.dispose()
        self.observer = None

        if self.state is State.WATCHING:
            self.instance.end_watch_event_logs(self.run_id, self.handle_new_event)
        elif self.state is State.LOADING:
            self.stopping.set()

    def handle_new_event(self, new_event):
        if self.observer:
            self.observer.on_next(([new_event], False))



import graphene

from ..errors import GrapheneAssetNotFoundError, GraphenePythonError
from ..pipelines.pipeline import GrapheneAsset
from ..util import non_null_list


class GrapheneAssetConnection(graphene.ObjectType):
    nodes = non_null_list(GrapheneAsset)

    class Meta:
        name = "AssetConnection"


class GrapheneAssetsOrError(graphene.Union):
    class Meta:
        types = (GrapheneAssetConnection, GraphenePythonError)
        name = "AssetsOrError"


class GrapheneAssetOrError(graphene.Union):
    class Meta:
        types = (GrapheneAsset, GrapheneAssetNotFoundError)
        name = "AssetOrError"



import json

from dagster import check
from dagster.config.config_type import ALL_CONFIG_BUILTINS
from dagster.utils import file_relative_path
from dagster_graphql.test.utils import execute_dagster_graphql, infer_pipeline_selector

from .graphql_context_test_suite import NonLaunchableGraphQLContextTestMatrix
from .setup import csv_hello_world_solids_config

CONFIG_VALIDATION_QUERY = """
query PipelineQuery(
    $runConfigData: RunConfigData,
    $pipeline: PipelineSelector!,
    $mode: String!
) {
    isPipelineConfigValid(
        runConfigData: $runConfigData,
        pipeline: $pipeline,
        mode: $mode
    ) {
        __typename
        ... on PipelineConfigValidationValid {
            pipelineName
        }
        ... on RunConfigValidationInvalid {
            pipelineName
            errors {
                __typename
                ... on RuntimeMismatchConfigError {
                    valueRep
                }
                ... on MissingFieldConfigError {
                    field { name }
                }
                ... on MissingFieldsConfigError {
                    fields { name }
                }
                ... on FieldNotDefinedConfigError {
                    fieldName
                }
                ... on FieldsNotDefinedConfigError {
                    fieldNames
                }
                ... on SelectorTypeConfigError {
                    incomingFields
                }
                message
                reason
                stack {
                    entries {
                        __typename
                        ... on EvaluationStackPathEntry {
                            fieldName
                        }
                        ... on EvaluationStackListItemEntry {
                            listIndex
                        }
                    }
                }
            }
        }
        ... on PipelineNotFoundError {
            pipelineName
        }
        ... on PythonError {
            message
            stack
        }
    }
}
"""


def field_stack(error_data):
    return [
        entry["fieldName"]
        for entry in error_data["stack"]["entries"]
        if entry["__typename"] == "EvaluationStackPathEntry"
    ]


def single_error_data(result):
    assert len(result.data["isPipelineConfigValid"]["errors"]) == 1
    return result.data["isPipelineConfigValid"]["errors"][0]


def find_error(result, field_stack_to_find, reason):
    llist = list(find_errors(result, field_stack_to_find, reason))
    assert len(llist) == 1
    return llist[0]


def find_errors(result, field_stack_to_find, reason):
    error_datas = result.data["isPipelineConfigValid"]["errors"]
    for error_data in error_datas:
        if field_stack_to_find == field_stack(error_data) and error_data["reason"] == reason:
            yield error_data


def execute_config_graphql(context, pipeline_name, run_config, mode):
    selector = infer_pipeline_selector(context, pipeline_name)
    return execute_dagster_graphql(
        context,
        CONFIG_VALIDATION_QUERY,
        {
            "runConfigData": run_config,
            "pipeline": selector,
            "mode": mode,
        },
    )


class TestConfigTypes(NonLaunchableGraphQLContextTestMatrix):
    def test_pipeline_not_found(self, graphql_context):
        result = execute_config_graphql(
            graphql_context, pipeline_name="nope", run_config={}, mode="default"
        )

        assert not result.errors
        assert result.data
        assert result.data["isPipelineConfigValid"]["__typename"] == "PipelineNotFoundError"
        assert result.data["isPipelineConfigValid"]["pipelineName"] == "nope"

    def test_basic_valid_config(self, graphql_context):
        result = execute_config_graphql(
            graphql_context,
            pipeline_name="csv_hello_world",
            run_config=csv_hello_world_solids_config(),
            mode="default",
        )

        assert not result.errors
        assert result.data
        assert result.data["isPipelineConfigValid"]["__typename"] == "PipelineConfigValidationValid"
        assert result.data["isPipelineConfigValid"]["pipelineName"] == "csv_hello_world"

    def test_basic_valid_config_serialized_config(self, graphql_context):
        result = execute_config_graphql(
            graphql_context,
            pipeline_name="csv_hello_world",
            run_config=json.dumps(csv_hello_world_solids_config()),
            mode="default",
        )

        assert not result.errors
        assert result.data
        assert result.data["isPipelineConfigValid"]["__typename"] == "PipelineConfigValidationValid"
        assert result.data["isPipelineConfigValid"]["pipelineName"] == "csv_hello_world"

    def test_root_field_not_defined(self, graphql_context):
        result = execute_config_graphql(
            graphql_context,
            pipeline_name="csv_hello_world",
            run_config={
                "solids": {
                    "sum_solid": {
                        "inputs": {"num": file_relative_path(__file__, "../data/num.csv")}
                    }
                },
                "nope": {},
            },
            mode="default",
        )

        assert not result.errors
        assert result.data
        assert result.data["isPipelineConfigValid"]["__typename"] == "RunConfigValidationInvalid"
        assert result.data["isPipelineConfigValid"]["pipelineName"] == "csv_hello_world"
        errors = result.data["isPipelineConfigValid"]["errors"]
        assert len(errors) == 1
        error = errors[0]
        assert error["__typename"] == "FieldNotDefinedConfigError"
        assert error["fieldName"] == "nope"
        assert not error["stack"]["entries"]

    def test_basic_invalid_not_defined_field(self, graphql_context):
        result = execute_config_graphql(
            graphql_context,
            pipeline_name="csv_hello_world",
            run_config={"solids": {"sum_solid": {"inputs": {"num": "foo.txt", "extra": "nope"}}}},
            mode="default",
        )

        assert not result.errors
        assert result.data
        assert result.data["isPipelineConfigValid"]["__typename"] == "RunConfigValidationInvalid"
        assert result.data["isPipelineConfigValid"]["pipelineName"] == "csv_hello_world"
        assert len(result.data["isPipelineConfigValid"]["errors"]) == 1
        error_data = result.data["isPipelineConfigValid"]["errors"][0]
        assert ["solids", "sum_solid", "inputs"] == field_stack(error_data)
        assert error_data["reason"] == "FIELD_NOT_DEFINED"
        assert error_data["fieldName"] == "extra"

    def test_multiple_not_defined_fields(self, graphql_context):
        result = execute_config_graphql(
            graphql_context,
            pipeline_name="csv_hello_world",
            run_config={
                "solids": {
                    "sum_solid": {
                        "inputs": {"num": "foo.txt", "extra_one": "nope", "extra_two": "nope"}
                    }
                }
            },
            mode="default",
        )

        assert not result.errors
        assert result.data
        assert result.data["isPipelineConfigValid"]["__typename"] == "RunConfigValidationInvalid"
        assert result.data["isPipelineConfigValid"]["pipelineName"] == "csv_hello_world"
        assert len(result.data["isPipelineConfigValid"]["errors"]) == 1
        error_data = result.data["isPipelineConfigValid"]["errors"][0]
        assert ["solids", "sum_solid", "inputs"] == field_stack(error_data)
        assert error_data["reason"] == "FIELDS_NOT_DEFINED"
        assert error_data["fieldNames"] == ["extra_one", "extra_two"]

    def test_root_wrong_type(self, graphql_context):
        result = execute_config_graphql(
            graphql_context, pipeline_name="csv_hello_world", run_config=123, mode="default"
        )

        assert not result.errors
        assert result.data
        assert result.data["isPipelineConfigValid"]["__typename"] == "RunConfigValidationInvalid"
        assert result.data["isPipelineConfigValid"]["pipelineName"] == "csv_hello_world"
        assert len(result.data["isPipelineConfigValid"]["errors"]) == 1
        error_data = result.data["isPipelineConfigValid"]["errors"][0]
        assert error_data["reason"] == "RUNTIME_TYPE_MISMATCH"

    def test_basic_invalid_config_type_mismatch(self, graphql_context):
        result = execute_config_graphql(
            graphql_context,
            pipeline_name="csv_hello_world",
            run_config={"solids": {"sum_solid": {"inputs": {"num": 123}}}},
            mode="default",
        )

        assert not result.errors
        assert result.data
        assert result.data["isPipelineConfigValid"]["__typename"] == "RunConfigValidationInvalid"
        assert result.data["isPipelineConfigValid"]["pipelineName"] == "csv_hello_world"
        assert len(result.data["isPipelineConfigValid"]["errors"]) == 1
        error_data = result.data["isPipelineConfigValid"]["errors"][0]
        assert error_data["message"]
        assert error_data["stack"]
        assert error_data["stack"]["entries"]
        assert error_data["reason"] == "RUNTIME_TYPE_MISMATCH"
        assert error_data["valueRep"] == "123"

        assert ["solids", "sum_solid", "inputs", "num"] == field_stack(error_data)

    def test_basic_invalid_config_missing_field(self, graphql_context):
        result = execute_config_graphql(
            graphql_context,
            pipeline_name="csv_hello_world",
            run_config={"solids": {"sum_solid": {"inputs": {}}}},
            mode="default",
        )

        assert not result.errors
        assert result.data
        assert result.data["isPipelineConfigValid"]["__typename"] == "RunConfigValidationInvalid"
        assert result.data["isPipelineConfigValid"]["pipelineName"] == "csv_hello_world"
        assert len(result.data["isPipelineConfigValid"]["errors"]) == 1
        error_data = result.data["isPipelineConfigValid"]["errors"][0]

        assert ["solids", "sum_solid", "inputs"] == field_stack(error_data)
        assert error_data["reason"] == "MISSING_REQUIRED_FIELD"
        assert error_data["field"]["name"] == "num"

    def test_mode_resource_config_works(self, graphql_context):
        result = execute_config_graphql(
            graphql_context,
            pipeline_name="multi_mode_with_resources",
            run_config={"resources": {"op": {"config": 2}}},
            mode="add_mode",
        )

        assert not result.errors
        assert result.data
        assert result.data["isPipelineConfigValid"]["__typename"] == "PipelineConfigValidationValid"
        assert result.data["isPipelineConfigValid"]["pipelineName"] == "multi_mode_with_resources"

        result = execute_config_graphql(
            graphql_context,
            pipeline_name="multi_mode_with_resources",
            run_config={"resources": {"op": {"config": 2}}},
            mode="mult_mode",
        )

        assert not result.errors
        assert result.data
        assert result.data["isPipelineConfigValid"]["__typename"] == "PipelineConfigValidationValid"
        assert result.data["isPipelineConfigValid"]["pipelineName"] == "multi_mode_with_resources"

        result = execute_config_graphql(
            graphql_context,
            pipeline_name="multi_mode_with_resources",
            run_config={"resources": {"op": {"config": {"num_one": 2, "num_two": 3}}}},
            mode="double_adder",
        )

        assert not result.errors
        assert result.data
        assert result.data["isPipelineConfigValid"]["__typename"] == "PipelineConfigValidationValid"
        assert result.data["isPipelineConfigValid"]["pipelineName"] == "multi_mode_with_resources"

    def test_missing_resource(self, graphql_context):
        result = execute_config_graphql(
            graphql_context,
            pipeline_name="multi_mode_with_resources",
            run_config={"resources": {}},
            mode="add_mode",
        )

        assert not result.errors
        assert result.data
        assert result.data["isPipelineConfigValid"]["__typename"] == "RunConfigValidationInvalid"
        error_data = single_error_data(result)
        assert error_data["reason"] == "MISSING_REQUIRED_FIELD"
        assert error_data["field"]["name"] == "op"

    def test_undefined_resource(self, graphql_context):
        result = execute_config_graphql(
            graphql_context,
            pipeline_name="multi_mode_with_resources",
            run_config={"resources": {"nope": {}}},
            mode="add_mode",
        )

        assert not result.errors
        assert result.data
        assert result.data["isPipelineConfigValid"]["__typename"] == "RunConfigValidationInvalid"
        assert {"FieldNotDefinedConfigError", "MissingFieldConfigError"} == {
            error_data["__typename"]
            for error_data in result.data["isPipelineConfigValid"]["errors"]
        }

    def test_more_complicated_works(self, graphql_context):
        result = execute_config_graphql(
            graphql_context,
            pipeline_name="more_complicated_nested_config",
            run_config={
                "solids": {
                    "a_solid_with_multilayered_config": {
                        "config": {
                            "field_any": {"123": 123},
                            "field_one": "foo.txt",
                            "field_two": "yup",
                            "field_three": "mmmhmmm",
                            "nested_field": {"field_four_str": "yaya", "field_five_int": 234},
                        }
                    }
                }
            },
            mode="default",
        )
        assert not result.errors
        assert result.data
        valid_data = result.data["isPipelineConfigValid"]
        assert valid_data["__typename"] == "PipelineConfigValidationValid"
        assert valid_data["pipelineName"] == "more_complicated_nested_config"

    def test_multiple_missing_fields(self, graphql_context):

        result = execute_config_graphql(
            graphql_context,
            pipeline_name="more_complicated_nested_config",
            run_config={"solids": {"a_solid_with_multilayered_config": {"config": {}}}},
            mode="default",
        )

        assert not result.errors
        assert result.data
        valid_data = result.data["isPipelineConfigValid"]

        assert valid_data["__typename"] == "RunConfigValidationInvalid"
        assert valid_data["pipelineName"] == "more_complicated_nested_config"
        assert len(valid_data["errors"]) == 1
        error_data = valid_data["errors"][0]
        missing_names = {field_data["name"] for field_data in error_data["fields"]}
        assert missing_names == {"nested_field", "field_one", "field_any"}
        assert field_stack(error_data) == ["solids", "a_solid_with_multilayered_config", "config"]

    def test_more_complicated_multiple_errors(self, graphql_context):
        result = execute_config_graphql(
            graphql_context,
            pipeline_name="more_complicated_nested_config",
            run_config={
                "solids": {
                    "a_solid_with_multilayered_config": {
                        "config": {
                            "field_any": [],
                            # 'field_one': 'foo.txt', # missing
                            "field_two": "yup",
                            "field_three": "mmmhmmm",
                            "extra_one": "kjsdkfjd",  # extra
                            "nested_field": {
                                "field_four_str": 23434,  # runtime type
                                "field_five_int": 234,
                                "extra_two": "ksjdkfjd",  # another extra
                            },
                        }
                    }
                }
            },
            mode="default",
        )

        assert not result.errors
        assert result.data
        valid_data = result.data["isPipelineConfigValid"]

        assert valid_data["__typename"] == "RunConfigValidationInvalid"
        assert valid_data["pipelineName"] == "more_complicated_nested_config"
        assert len(valid_data["errors"]) == 4

        missing_error_one = find_error(
            result,
            ["solids", "a_solid_with_multilayered_config", "config"],
            "MISSING_REQUIRED_FIELD",
        )
        assert ["solids", "a_solid_with_multilayered_config", "config"] == field_stack(
            missing_error_one
        )
        assert missing_error_one["reason"] == "MISSING_REQUIRED_FIELD"
        assert missing_error_one["field"]["name"] == "field_one"

        not_defined_one = find_error(
            result, ["solids", "a_solid_with_multilayered_config", "config"], "FIELD_NOT_DEFINED"
        )
        assert ["solids", "a_solid_with_multilayered_config", "config"] == field_stack(
            not_defined_one
        )
        assert not_defined_one["reason"] == "FIELD_NOT_DEFINED"
        assert not_defined_one["fieldName"] == "extra_one"

        dagster_type_error = find_error(
            result,
            [
                "solids",
                "a_solid_with_multilayered_config",
                "config",
                "nested_field",
                "field_four_str",
            ],
            "RUNTIME_TYPE_MISMATCH",
        )
        assert [
            "solids",
            "a_solid_with_multilayered_config",
            "config",
            "nested_field",
            "field_four_str",
        ] == field_stack(dagster_type_error)
        assert dagster_type_error["reason"] == "RUNTIME_TYPE_MISMATCH"
        assert dagster_type_error["valueRep"] == "23434"

        not_defined_two = find_error(
            result,
            ["solids", "a_solid_with_multilayered_config", "config", "nested_field"],
            "FIELD_NOT_DEFINED",
        )

        assert [
            "solids",
            "a_solid_with_multilayered_config",
            "config",
            "nested_field",
        ] == field_stack(not_defined_two)
        assert not_defined_two["reason"] == "FIELD_NOT_DEFINED"
        assert not_defined_two["fieldName"] == "extra_two"

        # TODO: two more errors

    def test_config_list(self, graphql_context):
        result = execute_config_graphql(
            graphql_context,
            pipeline_name="pipeline_with_list",
            run_config={"solids": {"solid_with_list": {"config": [1, 2]}}},
            mode="default",
        )

        assert not result.errors
        assert result.data
        valid_data = result.data["isPipelineConfigValid"]
        assert valid_data["__typename"] == "PipelineConfigValidationValid"
        assert valid_data["pipelineName"] == "pipeline_with_list"

    def test_config_list_invalid(self, graphql_context):
        result = execute_config_graphql(
            graphql_context,
            pipeline_name="pipeline_with_list",
            run_config={"solids": {"solid_with_list": {"config": "foo"}}},
            mode="default",
        )

        assert not result.errors
        assert result.data
        valid_data = result.data["isPipelineConfigValid"]
        assert valid_data["__typename"] == "RunConfigValidationInvalid"
        assert valid_data["pipelineName"] == "pipeline_with_list"
        assert len(valid_data["errors"]) == 1
        assert ["solids", "solid_with_list", "config"] == field_stack(valid_data["errors"][0])

    def test_config_list_item_invalid(self, graphql_context):
        result = execute_config_graphql(
            graphql_context,
            pipeline_name="pipeline_with_list",
            run_config={"solids": {"solid_with_list": {"config": [1, "foo"]}}},
            mode="default",
        )

        assert not result.errors
        assert result.data
        valid_data = result.data["isPipelineConfigValid"]
        assert valid_data["__typename"] == "RunConfigValidationInvalid"
        assert valid_data["pipelineName"] == "pipeline_with_list"
        assert len(valid_data["errors"]) == 1
        entries = valid_data["errors"][0]["stack"]["entries"]
        assert len(entries) == 4
        assert ["solids", "solid_with_list", "config"] == field_stack(valid_data["errors"][0])

        last_entry = entries[3]
        assert last_entry["__typename"] == "EvaluationStackListItemEntry"
        assert last_entry["listIndex"] == 1

    def test_smoke_test_config_type_system(self, graphql_context):
        selector = infer_pipeline_selector(graphql_context, "more_complicated_nested_config")
        result = execute_dagster_graphql(
            graphql_context,
            ALL_CONFIG_TYPES_QUERY,
            {"selector": selector, "mode": "default"},
        )

        config_types_data = result.data["runConfigSchemaOrError"]["allConfigTypes"]

        assert has_config_type_with_key_prefix(config_types_data, "Shape.")

        for builtin_config_type in ALL_CONFIG_BUILTINS:
            assert has_config_type(config_types_data, builtin_config_type.given_name)


def pipeline_named(result, name):
    for pipeline_data in result.data["pipelines"]["nodes"]:
        if pipeline_data["name"] == name:
            return pipeline_data
    check.failed("Did not find")


def has_config_type_with_key_prefix(config_types_data, prefix):
    for config_type_data in config_types_data:
        if config_type_data["key"].startswith(prefix):
            return True

    return False


def has_config_type(config_types_data, name):
    for config_type_data in config_types_data:
        if config_type_data.get("givenName") == name:
            return True

    return False


ALL_CONFIG_TYPES_QUERY = """
fragment configTypeFragment on ConfigType {
  __typename
  key
  description
  isSelector
  typeParamKeys
  recursiveConfigTypes {
    key
    description
    ... on CompositeConfigType {
        fields {
            name
            isRequired
            description
        }
    }
    ... on WrappingConfigType {
        ofType { key }
    }
  }
  ... on EnumConfigType {
    givenName
    values {
      value
      description
    }
  }
  ... on RegularConfigType {
    givenName
  }
  ... on CompositeConfigType {
    fields {
      name
      isRequired
      description
    }
  }
  ... on WrappingConfigType {
    ofType { key }
  }
  ... on ScalarUnionConfigType {
    scalarType { key }
    nonScalarType { key }
  }
}

query allConfigTypes($selector: PipelineSelector!, $mode: String!) {
  runConfigSchemaOrError(selector: $selector, mode: $mode ) {
    ... on RunConfigSchema {
      allConfigTypes {
        ...configTypeFragment
      }
    }
  }
}
"""


def get_field_data(config_type_data, name):
    for field_data in config_type_data["fields"]:
        if field_data["name"] == name:
            return field_data


def get_field_names(config_type_data):
    return {field_data["name"] for field_data in config_type_data.get("fields", [])}



from unittest.mock import Mock

import pytest
from dagster import check
from dagster.core.workspace.permissions import EDITOR_PERMISSIONS, VIEWER_PERMISSIONS, Permissions
from dagster_graphql.implementation.utils import (
    UserFacingGraphQLError,
    assert_permission,
    check_permission,
)
from dagster_graphql.test.utils import execute_dagster_graphql

from .graphql_context_test_suite import NonLaunchableGraphQLContextTestMatrix

PERMISSIONS_QUERY = """
    query PermissionsQuery {
      permissions {
        permission
        value
      }
    }
"""


class FakeMutation:
    @check_permission("fake_permission")
    def mutate(self, graphene_info, **_kwargs):
        pass


class FakeOtherPermissionMutation:
    @check_permission("fake_other_permission")
    def mutate(self, graphene_info, **_kwargs):
        pass


class FakeMissingPermissionMutation:
    @check_permission("fake_missing_permission")
    def mutate(self, graphene_info, **_kwargs):
        pass


class FakeEnumPermissionMutation:
    @check_permission(Permissions.LAUNCH_PIPELINE_EXECUTION)
    def mutate(self, graphene_info, **_kwargs):
        pass


class FakeOtherEnumPermisisonMutation:
    @check_permission(Permissions.LAUNCH_PIPELINE_REEXECUTION)
    def mutate(self, graphene_info, **_kwargs):
        pass


class FakeMissingEnumPermisisonMutation:
    @check_permission(Permissions.LAUNCH_PARTITION_BACKFILL)
    def mutate(self, graphene_info, **_kwargs):
        pass


@pytest.fixture(name="fake_graphene_info")
def fake_graphene_info_fixture():
    context = Mock()

    def fake_has_permission(permission):
        permission_map = {
            "fake_permission": True,
            "fake_other_permission": False,
            Permissions.LAUNCH_PIPELINE_EXECUTION: True,
            Permissions.LAUNCH_PIPELINE_REEXECUTION: False,
        }
        check.invariant(permission in permission_map, "Permission does not exist in map")
        return permission_map[permission]

    context.has_permission = Mock(side_effect=fake_has_permission)

    graphene_info = Mock()
    graphene_info.context = context

    return graphene_info


@pytest.mark.parametrize("mutation", [FakeMutation(), FakeEnumPermissionMutation()])
def test_check_permission_has_permission(fake_graphene_info, mutation):
    mutation.mutate(fake_graphene_info)


@pytest.mark.parametrize(
    "mutation", [FakeOtherPermissionMutation(), FakeOtherEnumPermisisonMutation()]
)
def test_check_permission_does_not_have_permission(fake_graphene_info, mutation):
    with pytest.raises(UserFacingGraphQLError, match="GrapheneUnauthorizedError"):
        mutation.mutate(fake_graphene_info)


@pytest.mark.parametrize(
    "mutation", [FakeMissingPermissionMutation(), FakeMissingEnumPermisisonMutation()]
)
def test_check_permission_permission_does_not_exist(fake_graphene_info, mutation):
    with pytest.raises(check.CheckError):
        mutation.mutate(fake_graphene_info)


@pytest.mark.parametrize("permission", ["fake_permission", Permissions.LAUNCH_PIPELINE_EXECUTION])
def test_assert_permission_has_permission(fake_graphene_info, permission):
    assert_permission(fake_graphene_info, permission)


@pytest.mark.parametrize(
    "permission", ["fake_other_permission", Permissions.LAUNCH_PIPELINE_REEXECUTION]
)
def test_assert_permission_does_not_have_permission(fake_graphene_info, permission):
    with pytest.raises(UserFacingGraphQLError, match="GrapheneUnauthorizedError"):
        assert_permission(fake_graphene_info, permission)


@pytest.mark.parametrize(
    "permission", ["fake_missing_permission", Permissions.LAUNCH_PARTITION_BACKFILL]
)
def test_assert_permission_permission_does_not_exist(fake_graphene_info, permission):
    with pytest.raises(check.CheckError):
        assert_permission(fake_graphene_info, permission)


class TestPermissionsQuery(NonLaunchableGraphQLContextTestMatrix):
    def test_permissions_query(self, graphql_context):
        result = execute_dagster_graphql(graphql_context, PERMISSIONS_QUERY)
        assert result.data

        assert result.data["permissions"]

        permissions_map = {
            permission["permission"]: permission["value"]
            for permission in result.data["permissions"]
        }

        if graphql_context.read_only:
            assert permissions_map == VIEWER_PERMISSIONS
        else:
            assert permissions_map == EDITOR_PERMISSIONS



import graphene
from dagster import check
from dagster.core.host_representation import ExternalSensor, ExternalTargetData, SensorSelector
from dagster.core.scheduler.instigation import InstigatorState
from dagster.core.workspace.permissions import Permissions
from dagster_graphql.implementation.utils import capture_error, check_permission

from ..implementation.fetch_sensors import get_sensor_next_tick, start_sensor, stop_sensor
from .asset_key import GrapheneAssetKey
from .errors import (
    GraphenePythonError,
    GrapheneRepositoryNotFoundError,
    GrapheneSensorNotFoundError,
    GrapheneUnauthorizedError,
)
from .inputs import GrapheneSensorSelector
from .instigation import GrapheneFutureInstigationTick, GrapheneInstigationState
from .util import non_null_list


class GrapheneTarget(graphene.ObjectType):
    pipelineName = graphene.NonNull(graphene.String)
    mode = graphene.NonNull(graphene.String)
    solidSelection = graphene.List(graphene.NonNull(graphene.String))

    class Meta:
        name = "Target"

    def __init__(self, external_target):
        self._external_target = check.inst_param(
            external_target, "external_target", ExternalTargetData
        )
        super().__init__(
            pipelineName=external_target.pipeline_name,
            mode=external_target.mode,
            solidSelection=external_target.solid_selection,
        )


class GrapheneSensorMetadata(graphene.ObjectType):
    assetKeys = graphene.List(graphene.NonNull(GrapheneAssetKey))

    class Meta:
        name = "SensorMetadata"


class GrapheneSensor(graphene.ObjectType):
    id = graphene.NonNull(graphene.ID)
    jobOriginId = graphene.NonNull(graphene.String)
    name = graphene.NonNull(graphene.String)
    targets = graphene.List(graphene.NonNull(GrapheneTarget))
    sensorState = graphene.NonNull(GrapheneInstigationState)
    minIntervalSeconds = graphene.NonNull(graphene.Int)
    description = graphene.String()
    nextTick = graphene.Field(GrapheneFutureInstigationTick)
    metadata = graphene.NonNull(GrapheneSensorMetadata)

    class Meta:
        name = "Sensor"

    def __init__(self, external_sensor, sensor_state):
        self._external_sensor = check.inst_param(external_sensor, "external_sensor", ExternalSensor)
        self._sensor_state = check.opt_inst_param(sensor_state, "sensor_state", InstigatorState)

        if not self._sensor_state:
            self._sensor_state = self._external_sensor.get_default_instigation_state()

        super().__init__(
            name=external_sensor.name,
            jobOriginId=external_sensor.get_external_origin_id(),
            minIntervalSeconds=external_sensor.min_interval_seconds,
            description=external_sensor.description,
            targets=[GrapheneTarget(target) for target in external_sensor.get_external_targets()],
            metadata=GrapheneSensorMetadata(
                assetKeys=external_sensor.metadata.asset_keys if external_sensor.metadata else None
            ),
        )

    def resolve_id(self, _):
        return self._external_sensor.get_external_origin_id()

    def resolve_sensorState(self, _graphene_info):
        return GrapheneInstigationState(self._sensor_state)

    def resolve_nextTick(self, graphene_info):
        return get_sensor_next_tick(graphene_info, self._sensor_state)


class GrapheneSensorOrError(graphene.Union):
    class Meta:
        types = (
            GrapheneSensor,
            GrapheneSensorNotFoundError,
            GrapheneUnauthorizedError,
            GraphenePythonError,
        )
        name = "SensorOrError"


class GrapheneSensors(graphene.ObjectType):
    results = non_null_list(GrapheneSensor)

    class Meta:
        name = "Sensors"


class GrapheneSensorsOrError(graphene.Union):
    class Meta:
        types = (GrapheneSensors, GrapheneRepositoryNotFoundError, GraphenePythonError)
        name = "SensorsOrError"


class GrapheneStartSensorMutation(graphene.Mutation):
    Output = graphene.NonNull(GrapheneSensorOrError)

    class Arguments:
        sensor_selector = graphene.NonNull(GrapheneSensorSelector)

    class Meta:
        name = "StartSensorMutation"

    @capture_error
    @check_permission(Permissions.START_SENSOR)
    def mutate(self, graphene_info, sensor_selector):
        return start_sensor(graphene_info, SensorSelector.from_graphql_input(sensor_selector))


class GrapheneStopSensorMutationResult(graphene.ObjectType):
    instigationState = graphene.Field(GrapheneInstigationState)

    class Meta:
        name = "StopSensorMutationResult"

    def __init__(self, job_state):
        super().__init__()
        self._job_state = check.inst_param(job_state, "job_state", InstigatorState)

    def resolve_instigationState(self, _graphene_info):
        if not self._job_state:
            return None

        return GrapheneInstigationState(job_state=self._job_state)


class GrapheneStopSensorMutationResultOrError(graphene.Union):
    class Meta:
        types = (GrapheneStopSensorMutationResult, GrapheneUnauthorizedError, GraphenePythonError)
        name = "StopSensorMutationResultOrError"


class GrapheneStopSensorMutation(graphene.Mutation):
    Output = graphene.NonNull(GrapheneStopSensorMutationResultOrError)

    class Arguments:
        job_origin_id = graphene.NonNull(graphene.String)

    class Meta:
        name = "StopSensorMutation"

    @capture_error
    @check_permission(Permissions.STOP_SENSOR)
    def mutate(self, graphene_info, job_origin_id):
        return stop_sensor(graphene_info, job_origin_id)


types = [
    GrapheneSensor,
    GrapheneSensorOrError,
    GrapheneSensors,
    GrapheneSensorsOrError,
    GrapheneStopSensorMutation,
    GrapheneStopSensorMutationResult,
    GrapheneStopSensorMutationResultOrError,
    GrapheneStopSensorMutation,
]



from typing import Dict

from setuptools import find_packages, setup  # type: ignore


def get_version() -> str:
    version: Dict[str, str] = {}
    with open("dagster_graphql/version.py") as fp:
        exec(fp.read(), version)  # pylint: disable=W0122

    return version["__version__"]


if __name__ == "__main__":
    ver = get_version()
    # dont pin dev installs to avoid pip dep resolver issues
    pin = "" if ver == "dev" else f"=={ver}"
    setup(
        name="dagster-graphql",
        version=ver,
        author="Elementl",
        author_email="hello@elementl.com",
        license="Apache-2.0",
        description="The GraphQL frontend to python dagster.",
        url="https://github.com/dagster-io/dagster/tree/master/python_modules/dagster-graphql",
        classifiers=[
            "Programming Language :: Python :: 3.6",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "License :: OSI Approved :: Apache Software License",
            "Operating System :: OS Independent",
        ],
        packages=find_packages(exclude=["test"]),
        install_requires=[
            f"dagster{pin}",
            "graphene>=2.1.3,<3",  # compatability with graphql-ws in dagit
            "graphql-core>=2.1,<3",  # compatability with graphql-ws in dagit
            "gevent-websocket>=0.10.1",
            "gevent",
            "requests",
            "gql<3",
        ],
        entry_points={"console_scripts": ["dagster-graphql = dagster_graphql.cli:main"]},
    )



import graphene
import yaml
from dagster import check
from dagster.core.events import AssetKey, StepMaterializationData
from dagster.core.events.log import EventLogEntry
from dagster.core.host_representation.external import ExternalExecutionPlan, ExternalPipeline
from dagster.core.host_representation.external_data import ExternalPresetData
from dagster.core.storage.pipeline_run import (
    PipelineRun,
    PipelineRunStatus,
    PipelineRunsFilter,
    RunRecord,
)
from dagster.core.storage.tags import TagType, get_tag_type

from ...implementation.events import construct_basic_params, from_event_record
from ...implementation.fetch_assets import get_assets_for_run_id
from ...implementation.fetch_pipelines import get_pipeline_reference_or_raise
from ...implementation.fetch_runs import get_run_by_id, get_runs, get_stats, get_step_stats
from ...implementation.fetch_schedules import get_schedules_for_pipeline
from ...implementation.fetch_sensors import get_sensors_for_pipeline
from ...implementation.utils import UserFacingGraphQLError, capture_error
from ..asset_key import GrapheneAssetKey
from ..dagster_types import GrapheneDagsterType, GrapheneDagsterTypeOrError, to_dagster_type
from ..errors import GrapheneDagsterTypeNotFoundError, GraphenePythonError, GrapheneRunNotFoundError
from ..execution import GrapheneExecutionPlan
from ..inputs import GrapheneAssetKeyInput
from ..logs.compute_logs import GrapheneComputeLogs
from ..logs.events import (
    GrapheneDagsterRunEvent,
    GrapheneRunStepStats,
    GrapheneStepMaterializationEvent,
)
from ..paging import GrapheneCursor
from ..repository_origin import GrapheneRepositoryOrigin
from ..runs import GrapheneRunConfigData
from ..schedules.schedules import GrapheneSchedule
from ..sensors import GrapheneSensor
from ..solids import (
    GrapheneSolid,
    GrapheneSolidContainer,
    GrapheneSolidHandle,
    build_solid_handles,
    build_solids,
)
from ..tags import GraphenePipelineTag
from ..util import non_null_list
from .mode import GrapheneMode
from .pipeline_ref import GraphenePipelineReference
from .pipeline_run_stats import GrapheneRunStatsSnapshotOrError
from .status import GrapheneRunStatus

STARTED_STATUSES = {
    PipelineRunStatus.STARTED,
    PipelineRunStatus.SUCCESS,
    PipelineRunStatus.FAILURE,
    PipelineRunStatus.CANCELED,
}

COMPLETED_STATUSES = {
    PipelineRunStatus.FAILURE,
    PipelineRunStatus.SUCCESS,
    PipelineRunStatus.CANCELED,
}


class GrapheneAssetMaterialization(graphene.ObjectType):
    materializationEvent = graphene.NonNull(GrapheneStepMaterializationEvent)
    runOrError = graphene.NonNull(lambda: GrapheneRunOrError)
    partition = graphene.Field(graphene.String)

    class Meta:
        name = "AssetMaterialization"

    def __init__(self, event):
        super().__init__()
        self._event = check.inst_param(event, "event", EventLogEntry)
        check.invariant(
            isinstance(event.dagster_event.step_materialization_data, StepMaterializationData)
        )

    def resolve_materializationEvent(self, _graphene_info):
        return GrapheneStepMaterializationEvent(
            materialization=self._event.dagster_event.step_materialization_data.materialization,
            assetLineage=self._event.dagster_event.step_materialization_data.asset_lineage,
            **construct_basic_params(self._event),
        )

    def resolve_runOrError(self, graphene_info):
        return get_run_by_id(graphene_info, self._event.run_id)

    def resolve_partition(self, _graphene_info):
        return self._event.dagster_event.step_materialization_data.materialization.partition


class GrapheneMaterializationCount(graphene.ObjectType):
    partition = graphene.NonNull(graphene.String)
    materializationCount = graphene.NonNull(graphene.Int)

    class Meta:
        name = "MaterializationCountByPartition"


class GrapheneAsset(graphene.ObjectType):
    id = graphene.NonNull(graphene.String)
    key = graphene.NonNull(GrapheneAssetKey)
    assetMaterializations = graphene.Field(
        non_null_list(GrapheneAssetMaterialization),
        partitions=graphene.List(graphene.String),
        partitionInLast=graphene.Int(),
        beforeTimestampMillis=graphene.String(),
        limit=graphene.Int(),
    )
    definition = graphene.Field("dagster_graphql.schema.asset_graph.GrapheneAssetNode")

    class Meta:
        name = "Asset"

    def __init__(self, key, definition=None):
        super().__init__(key=key, definition=definition)
        self._definition = definition

    def resolve_id(self, _):
        return self.key

    def resolve_assetMaterializations(self, graphene_info, **kwargs):
        from ...implementation.fetch_assets import get_asset_events

        try:
            before_timestamp = (
                int(kwargs.get("beforeTimestampMillis")) / 1000.0
                if kwargs.get("beforeTimestampMillis")
                else None
            )
        except ValueError:
            before_timestamp = None

        limit = kwargs.get("limit")
        partitions = kwargs.get("partitions")
        partitionInLast = kwargs.get("partitionInLast")
        if partitionInLast and self._definition:
            partitions = self._definition.get_partition_keys()[-int(partitionInLast) :]

        return [
            GrapheneAssetMaterialization(event=event)
            for event in get_asset_events(
                graphene_info,
                self.key,
                partitions=partitions,
                before_timestamp=before_timestamp,
                limit=limit,
            )
        ]


class GraphenePipelineRun(graphene.Interface):
    id = graphene.NonNull(graphene.ID)
    runId = graphene.NonNull(graphene.String)
    # Nullable because of historical runs
    pipelineSnapshotId = graphene.String()
    repositoryOrigin = graphene.Field(GrapheneRepositoryOrigin)
    status = graphene.NonNull(GrapheneRunStatus)
    pipeline = graphene.NonNull(GraphenePipelineReference)
    pipelineName = graphene.NonNull(graphene.String)
    jobName = graphene.NonNull(graphene.String)
    solidSelection = graphene.List(graphene.NonNull(graphene.String))
    stats = graphene.NonNull(GrapheneRunStatsSnapshotOrError)
    stepStats = non_null_list(GrapheneRunStepStats)
    computeLogs = graphene.Field(
        graphene.NonNull(GrapheneComputeLogs),
        stepKey=graphene.Argument(graphene.NonNull(graphene.String)),
        description="""
        Compute logs are the stdout/stderr logs for a given solid step computation
        """,
    )
    executionPlan = graphene.Field(GrapheneExecutionPlan)
    stepKeysToExecute = graphene.List(graphene.NonNull(graphene.String))
    runConfigYaml = graphene.NonNull(graphene.String)
    runConfig = graphene.NonNull(GrapheneRunConfigData)
    mode = graphene.NonNull(graphene.String)
    tags = non_null_list(GraphenePipelineTag)
    rootRunId = graphene.Field(graphene.String)
    parentRunId = graphene.Field(graphene.String)
    canTerminate = graphene.NonNull(graphene.Boolean)
    assets = non_null_list(GrapheneAsset)
    events = graphene.Field(
        non_null_list(GrapheneDagsterRunEvent),
        after=graphene.Argument(GrapheneCursor),
    )

    class Meta:
        name = "PipelineRun"


class GrapheneRun(graphene.ObjectType):
    id = graphene.NonNull(graphene.ID)
    runId = graphene.NonNull(graphene.String)
    # Nullable because of historical runs
    pipelineSnapshotId = graphene.String()
    repositoryOrigin = graphene.Field(GrapheneRepositoryOrigin)
    status = graphene.NonNull(GrapheneRunStatus)
    pipeline = graphene.NonNull(GraphenePipelineReference)
    pipelineName = graphene.NonNull(graphene.String)
    jobName = graphene.NonNull(graphene.String)
    solidSelection = graphene.List(graphene.NonNull(graphene.String))
    resolvedOpSelection = graphene.List(graphene.NonNull(graphene.String))
    stats = graphene.NonNull(GrapheneRunStatsSnapshotOrError)
    stepStats = non_null_list(GrapheneRunStepStats)
    computeLogs = graphene.Field(
        graphene.NonNull(GrapheneComputeLogs),
        stepKey=graphene.Argument(graphene.NonNull(graphene.String)),
        description="""
        Compute logs are the stdout/stderr logs for a given solid step computation
        """,
    )
    executionPlan = graphene.Field(GrapheneExecutionPlan)
    stepKeysToExecute = graphene.List(graphene.NonNull(graphene.String))
    runConfigYaml = graphene.NonNull(graphene.String)
    runConfig = graphene.NonNull(GrapheneRunConfigData)
    mode = graphene.NonNull(graphene.String)
    tags = non_null_list(GraphenePipelineTag)
    rootRunId = graphene.Field(graphene.String)
    parentRunId = graphene.Field(graphene.String)
    canTerminate = graphene.NonNull(graphene.Boolean)
    assets = non_null_list(GrapheneAsset)
    events = graphene.Field(
        non_null_list(GrapheneDagsterRunEvent),
        after=graphene.Argument(GrapheneCursor),
    )
    startTime = graphene.Float()
    endTime = graphene.Float()

    class Meta:
        interfaces = (GraphenePipelineRun,)
        name = "Run"

    def __init__(self, run):
        pipeline_run = run.pipeline_run if isinstance(run, RunRecord) else run
        super().__init__(
            runId=pipeline_run.run_id,
            status=PipelineRunStatus(pipeline_run.status),
            mode=pipeline_run.mode,
        )
        self._pipeline_run = check.inst_param(pipeline_run, "pipeline_run", PipelineRun)
        self._run_record = run if isinstance(run, RunRecord) else None
        self._run_stats = None

    def resolve_id(self, _graphene_info):
        return self._pipeline_run.run_id

    def resolve_repositoryOrigin(self, _graphene_info):
        return (
            GrapheneRepositoryOrigin(
                self._pipeline_run.external_pipeline_origin.external_repository_origin
            )
            if self._pipeline_run.external_pipeline_origin
            else None
        )

    def resolve_pipeline(self, graphene_info):
        return get_pipeline_reference_or_raise(graphene_info, self._pipeline_run)

    def resolve_pipelineName(self, _graphene_info):
        return self._pipeline_run.pipeline_name

    def resolve_jobName(self, _graphene_info):
        return self._pipeline_run.pipeline_name

    def resolve_solidSelection(self, _graphene_info):
        return self._pipeline_run.solid_selection

    def resolve_resolvedOpSelection(self, _graphene_info):
        return self._pipeline_run.solids_to_execute

    def resolve_pipelineSnapshotId(self, _graphene_info):
        return self._pipeline_run.pipeline_snapshot_id

    def resolve_stats(self, graphene_info):
        return get_stats(graphene_info, self.run_id)

    def resolve_stepStats(self, graphene_info):
        return get_step_stats(graphene_info, self.run_id)

    def resolve_computeLogs(self, _graphene_info, stepKey):
        return GrapheneComputeLogs(runId=self.run_id, stepKey=stepKey)

    def resolve_executionPlan(self, graphene_info):
        if not (
            self._pipeline_run.execution_plan_snapshot_id
            and self._pipeline_run.pipeline_snapshot_id
        ):
            return None

        instance = graphene_info.context.instance

        execution_plan_snapshot = instance.get_execution_plan_snapshot(
            self._pipeline_run.execution_plan_snapshot_id
        )
        return (
            GrapheneExecutionPlan(
                ExternalExecutionPlan(execution_plan_snapshot=execution_plan_snapshot)
            )
            if execution_plan_snapshot
            else None
        )

    def resolve_stepKeysToExecute(self, _graphene_info):
        return self._pipeline_run.step_keys_to_execute

    def resolve_runConfigYaml(self, _graphene_info):
        return yaml.dump(
            self._pipeline_run.run_config, default_flow_style=False, allow_unicode=True
        )

    def resolve_runConfig(self, _graphene_info):
        return self._pipeline_run.run_config

    def resolve_tags(self, _graphene_info):
        return [
            GraphenePipelineTag(key=key, value=value)
            for key, value in self._pipeline_run.tags.items()
            if get_tag_type(key) != TagType.HIDDEN
        ]

    def resolve_rootRunId(self, _graphene_info):
        return self._pipeline_run.root_run_id

    def resolve_parentRunId(self, _graphene_info):
        return self._pipeline_run.parent_run_id

    @property
    def run_id(self):
        return self.runId

    def resolve_canTerminate(self, graphene_info):
        # short circuit if the pipeline run is in a terminal state
        if self._pipeline_run.is_finished:
            return False
        return graphene_info.context.instance.run_coordinator.can_cancel_run(self.run_id)

    def resolve_assets(self, graphene_info):
        return get_assets_for_run_id(graphene_info, self.run_id)

    def resolve_events(self, graphene_info, after=-1):
        events = graphene_info.context.instance.logs_after(self.run_id, cursor=after)
        return [from_event_record(event, self._pipeline_run.pipeline_name) for event in events]

    def _get_run_record(self, instance):
        if not self._run_record:
            self._run_record = instance.get_run_records(run_ids=[self.runId])[0]
        return self._run_record

    def resolve_startTime(self, graphene_info):
        run_record = self._get_run_record(graphene_info.context.instance)
        # If a user has not migrated in 0.13.15, then run_record will not have start_time and end_time. So it will be necessary to fill this data using the run_stats. Since we potentially make this call multiple times, we cache the result.
        if run_record.start_time is None and self._pipeline_run.status in STARTED_STATUSES:
            if self._run_stats is None or self._run_stats.start_time is None:
                self._run_stats = graphene_info.context.instance.get_run_stats(self.runId)
            return self._run_stats.start_time
        return run_record.start_time

    def resolve_endTime(self, graphene_info):
        run_record = self._get_run_record(graphene_info.context.instance)
        if run_record.end_time is None and self._pipeline_run.status in COMPLETED_STATUSES:
            if self._run_stats is None or self._run_stats.end_time is None:
                self._run_stats = graphene_info.context.instance.get_run_stats(self.runId)
            return self._run_stats.end_time
        return run_record.end_time


class GrapheneIPipelineSnapshotMixin:
    # Mixin this class to implement IPipelineSnapshot
    #
    # Graphene has some strange properties that make it so that you cannot
    # implement ABCs nor use properties in an overridable way. So the way
    # the mixin works is that the target classes have to have a method
    # get_represented_pipeline()
    #
    name = graphene.NonNull(graphene.String)
    description = graphene.String()
    id = graphene.NonNull(graphene.ID)
    pipeline_snapshot_id = graphene.NonNull(graphene.String)
    dagster_types = non_null_list(GrapheneDagsterType)
    dagster_type_or_error = graphene.Field(
        graphene.NonNull(GrapheneDagsterTypeOrError),
        dagsterTypeName=graphene.Argument(graphene.NonNull(graphene.String)),
    )
    solids = non_null_list(GrapheneSolid)
    modes = non_null_list(GrapheneMode)
    solid_handles = graphene.Field(
        non_null_list(GrapheneSolidHandle), parentHandleID=graphene.String()
    )
    solid_handle = graphene.Field(
        GrapheneSolidHandle,
        handleID=graphene.Argument(graphene.NonNull(graphene.String)),
    )
    tags = non_null_list(GraphenePipelineTag)
    runs = graphene.Field(
        non_null_list(GrapheneRun),
        cursor=graphene.String(),
        limit=graphene.Int(),
    )
    schedules = non_null_list(GrapheneSchedule)
    sensors = non_null_list(GrapheneSensor)
    parent_snapshot_id = graphene.String()
    graph_name = graphene.NonNull(graphene.String)

    class Meta:
        name = "IPipelineSnapshotMixin"

    def get_represented_pipeline(self):
        raise NotImplementedError()

    def resolve_pipeline_snapshot_id(self, _graphene_info):
        return self.get_represented_pipeline().identifying_pipeline_snapshot_id

    def resolve_id(self, _graphene_info):
        return self.get_represented_pipeline().identifying_pipeline_snapshot_id

    def resolve_name(self, _graphene_info):
        return self.get_represented_pipeline().name

    def resolve_description(self, _graphene_info):
        return self.get_represented_pipeline().description

    def resolve_dagster_types(self, _graphene_info):
        represented_pipeline = self.get_represented_pipeline()
        return sorted(
            list(
                map(
                    lambda dt: to_dagster_type(represented_pipeline.pipeline_snapshot, dt.key),
                    [t for t in represented_pipeline.dagster_type_snaps if t.name],
                )
            ),
            key=lambda dagster_type: dagster_type.name,
        )

    @capture_error
    def resolve_dagster_type_or_error(self, _graphene_info, **kwargs):
        type_name = kwargs["dagsterTypeName"]

        represented_pipeline = self.get_represented_pipeline()

        if not represented_pipeline.has_dagster_type_named(type_name):
            raise UserFacingGraphQLError(
                GrapheneDagsterTypeNotFoundError(dagster_type_name=type_name)
            )

        return to_dagster_type(
            represented_pipeline.pipeline_snapshot,
            represented_pipeline.get_dagster_type_by_name(type_name).key,
        )

    def resolve_solids(self, _graphene_info):
        represented_pipeline = self.get_represented_pipeline()
        return build_solids(
            represented_pipeline,
            represented_pipeline.dep_structure_index,
        )

    def resolve_modes(self, _graphene_info):
        represented_pipeline = self.get_represented_pipeline()
        return [
            GrapheneMode(
                represented_pipeline.config_schema_snapshot,
                represented_pipeline.identifying_pipeline_snapshot_id,
                mode_def_snap,
            )
            for mode_def_snap in sorted(
                represented_pipeline.mode_def_snaps, key=lambda item: item.name
            )
        ]

    def resolve_solid_handle(self, _graphene_info, handleID):
        return build_solid_handles(self.get_represented_pipeline()).get(handleID)

    def resolve_solid_handles(self, _graphene_info, **kwargs):
        handles = build_solid_handles(self.get_represented_pipeline())
        parentHandleID = kwargs.get("parentHandleID")

        if parentHandleID == "":
            handles = {key: handle for key, handle in handles.items() if not handle.parent}
        elif parentHandleID is not None:
            handles = {
                key: handle
                for key, handle in handles.items()
                if handle.parent and handle.parent.handleID.to_string() == parentHandleID
            }

        return [handles[key] for key in sorted(handles)]

    def resolve_tags(self, _graphene_info):
        represented_pipeline = self.get_represented_pipeline()
        return [
            GraphenePipelineTag(key=key, value=value)
            for key, value in represented_pipeline.pipeline_snapshot.tags.items()
        ]

    def resolve_solidSelection(self, _graphene_info):
        return self.get_represented_pipeline().solid_selection

    def resolve_runs(self, graphene_info, **kwargs):
        runs_filter = PipelineRunsFilter(pipeline_name=self.get_represented_pipeline().name)
        return get_runs(graphene_info, runs_filter, kwargs.get("cursor"), kwargs.get("limit"))

    def resolve_schedules(self, graphene_info):
        represented_pipeline = self.get_represented_pipeline()
        if not isinstance(represented_pipeline, ExternalPipeline):
            # this is an historical pipeline snapshot, so there are not any associated running
            # schedules
            return []

        pipeline_selector = represented_pipeline.handle.to_selector()
        schedules = get_schedules_for_pipeline(graphene_info, pipeline_selector)
        return schedules

    def resolve_sensors(self, graphene_info):
        represented_pipeline = self.get_represented_pipeline()
        if not isinstance(represented_pipeline, ExternalPipeline):
            # this is an historical pipeline snapshot, so there are not any associated running
            # sensors
            return []

        pipeline_selector = represented_pipeline.handle.to_selector()
        sensors = get_sensors_for_pipeline(graphene_info, pipeline_selector)
        return sensors

    def resolve_parent_snapshot_id(self, _graphene_info):
        lineage_snapshot = self.get_represented_pipeline().pipeline_snapshot.lineage_snapshot
        if lineage_snapshot:
            return lineage_snapshot.parent_snapshot_id
        else:
            return None

    def resolve_graph_name(self, _graphene_info):
        return self.get_represented_pipeline().get_graph_name()


class GrapheneIPipelineSnapshot(graphene.Interface):
    name = graphene.NonNull(graphene.String)
    description = graphene.String()
    pipeline_snapshot_id = graphene.NonNull(graphene.String)
    dagster_types = non_null_list(GrapheneDagsterType)
    dagster_type_or_error = graphene.Field(
        graphene.NonNull(GrapheneDagsterTypeOrError),
        dagsterTypeName=graphene.Argument(graphene.NonNull(graphene.String)),
    )
    solids = non_null_list(GrapheneSolid)
    modes = non_null_list(GrapheneMode)
    solid_handles = graphene.Field(
        non_null_list(GrapheneSolidHandle), parentHandleID=graphene.String()
    )
    solid_handle = graphene.Field(
        GrapheneSolidHandle,
        handleID=graphene.Argument(graphene.NonNull(graphene.String)),
    )
    tags = non_null_list(GraphenePipelineTag)
    runs = graphene.Field(
        non_null_list(GrapheneRun),
        cursor=graphene.String(),
        limit=graphene.Int(),
    )
    schedules = non_null_list(GrapheneSchedule)
    sensors = non_null_list(GrapheneSensor)
    parent_snapshot_id = graphene.String()
    graph_name = graphene.NonNull(graphene.String)

    class Meta:
        name = "IPipelineSnapshot"


class GraphenePipelinePreset(graphene.ObjectType):
    name = graphene.NonNull(graphene.String)
    solidSelection = graphene.List(graphene.NonNull(graphene.String))
    runConfigYaml = graphene.NonNull(graphene.String)
    mode = graphene.NonNull(graphene.String)
    tags = non_null_list(GraphenePipelineTag)

    class Meta:
        name = "PipelinePreset"

    def __init__(self, active_preset_data, pipeline_name):
        super().__init__()
        self._active_preset_data = check.inst_param(
            active_preset_data, "active_preset_data", ExternalPresetData
        )
        self._pipeline_name = check.str_param(pipeline_name, "pipeline_name")

    def resolve_name(self, _graphene_info):
        return self._active_preset_data.name

    def resolve_solidSelection(self, _graphene_info):
        return self._active_preset_data.solid_selection

    def resolve_runConfigYaml(self, _graphene_info):
        yaml_str = yaml.safe_dump(
            self._active_preset_data.run_config, default_flow_style=False, allow_unicode=True
        )
        return yaml_str if yaml_str else ""

    def resolve_mode(self, _graphene_info):
        return self._active_preset_data.mode

    def resolve_tags(self, _graphene_info):
        return [
            GraphenePipelineTag(key=key, value=value)
            for key, value in self._active_preset_data.tags.items()
            if get_tag_type(key) != TagType.HIDDEN
        ]


class GraphenePipeline(GrapheneIPipelineSnapshotMixin, graphene.ObjectType):
    id = graphene.NonNull(graphene.ID)
    presets = non_null_list(GraphenePipelinePreset)
    isJob = graphene.NonNull(graphene.Boolean)
    isAssetJob = graphene.NonNull(graphene.Boolean)
    repository = graphene.NonNull("dagster_graphql.schema.external.GrapheneRepository")
    assetNodes = graphene.Field(
        non_null_list("dagster_graphql.schema.asset_graph.GrapheneAssetNode"),
        assetKeys=graphene.Argument(graphene.List(graphene.NonNull(GrapheneAssetKeyInput))),
        loadMaterializations=graphene.Boolean(default_value=False),
    )

    class Meta:
        interfaces = (GrapheneSolidContainer, GrapheneIPipelineSnapshot)
        name = "Pipeline"

    def __init__(self, external_pipeline):
        super().__init__()
        self._external_pipeline = check.inst_param(
            external_pipeline, "external_pipeline", ExternalPipeline
        )

    def resolve_id(self, _graphene_info):
        return self._external_pipeline.get_external_origin_id()

    def get_represented_pipeline(self):
        return self._external_pipeline

    def resolve_presets(self, _graphene_info):
        return [
            GraphenePipelinePreset(preset, self._external_pipeline.name)
            for preset in sorted(self._external_pipeline.active_presets, key=lambda item: item.name)
        ]

    def resolve_isJob(self, _graphene_info):
        return self._external_pipeline.is_job

    def resolve_isAssetJob(self, graphene_info):
        handle = self._external_pipeline.repository_handle
        location = graphene_info.context.get_repository_location(handle.location_name)
        repository = location.get_repository(handle.repository_name)
        return bool(repository.get_external_asset_nodes(self._external_pipeline.name))

    def resolve_repository(self, graphene_info):
        from ..external import GrapheneRepository

        handle = self._external_pipeline.repository_handle
        location = graphene_info.context.get_repository_location(handle.location_name)
        return GrapheneRepository(location.get_repository(handle.repository_name), location)

    def resolve_assetNodes(self, graphene_info, **kwargs):
        from ..asset_graph import GrapheneAssetNode

        load_materializations = kwargs.get("loadMaterializations")

        handle = self._external_pipeline.repository_handle
        location = graphene_info.context.get_repository_location(handle.location_name)
        repository = location.get_repository(handle.repository_name)
        asset_nodes = repository.get_external_asset_nodes(self._external_pipeline.name)
        if not asset_nodes:
            return []
        asset_keys = set(
            AssetKey.from_graphql_input(asset_key) for asset_key in kwargs.get("assetKeys", [])
        )
        matching = [node for node in asset_nodes if not asset_keys or node.asset_key in asset_keys]
        if not matching:
            return []

        if load_materializations:
            events_by_key = graphene_info.context.instance.get_latest_materialization_events(
                [node.asset_key for node in matching]
            )
        else:
            events_by_key = {}

        return [
            GrapheneAssetNode(
                repository,
                node,
                events_by_key.get(node.asset_key),
                fetched_materialization=load_materializations,
            )
            for node in matching
        ]


class GrapheneJob(GraphenePipeline):
    class Meta:
        interfaces = (GrapheneSolidContainer, GrapheneIPipelineSnapshot)
        name = "Job"


class GrapheneGraph(graphene.ObjectType):
    class Meta:
        interfaces = (GrapheneSolidContainer,)
        name = "Graph"

    id = graphene.NonNull(graphene.ID)
    name = graphene.NonNull(graphene.String)
    description = graphene.String()
    solid_handle = graphene.Field(
        GrapheneSolidHandle,
        handleID=graphene.Argument(graphene.NonNull(graphene.String)),
    )
    solid_handles = graphene.Field(
        non_null_list(GrapheneSolidHandle), parentHandleID=graphene.String()
    )
    modes = non_null_list(GrapheneMode)

    def __init__(self, external_pipeline, solid_handle_id=None):
        self._external_pipeline = check.inst_param(
            external_pipeline, "external_pipeline", ExternalPipeline
        )
        self._solid_handle_id = check.opt_str_param(solid_handle_id, "solid_handle_id")
        super().__init__()

    def resolve_id(self, _graphene_info):
        if self._solid_handle_id:
            return (
                f"{self._external_pipeline.get_external_origin_id()}:solid:{self._solid_handle_id}"
            )
        return f"graph:{self._external_pipeline.get_external_origin_id()}"

    def resolve_name(self, _graphene_info):
        return self._external_pipeline.get_graph_name()

    def resolve_description(self, _graphene_info):
        return self._external_pipeline.description

    def resolve_solid_handle(self, _graphene_info, handleID):
        return build_solid_handles(self._external_pipeline).get(handleID)

    def resolve_solid_handles(self, _graphene_info, **kwargs):
        handles = build_solid_handles(self._external_pipeline)
        parentHandleID = kwargs.get("parentHandleID")

        if parentHandleID == "":
            handles = {key: handle for key, handle in handles.items() if not handle.parent}
        elif parentHandleID is not None:
            handles = {
                key: handle
                for key, handle in handles.items()
                if handle.parent and handle.parent.handleID.to_string() == parentHandleID
            }

        return [handles[key] for key in sorted(handles)]

    def resolve_modes(self, _graphene_info):
        # returns empty list... graphs don't have modes, this is a vestige of the old
        # pipeline explorer, which expected all solid containers to be pipelines
        return []


class GrapheneRunOrError(graphene.Union):
    class Meta:
        types = (GrapheneRun, GrapheneRunNotFoundError, GraphenePythonError)
        name = "RunOrError"


class GrapheneInProgressRunsByStep(graphene.ObjectType):
    stepKey = graphene.NonNull(graphene.String)
    unstartedRuns = non_null_list(GrapheneRun)
    inProgressRuns = non_null_list(GrapheneRun)

    class Meta:
        name = "InProgressRunsByStep"
